#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

/*
 * This is a CCP for frox. The sample code is licensed under the GPL
 * as seen in COPYING. This CCP tries to implement transparent
 * redirection of ftp clients to local mirrors. The mirrors are
 * defined in..., well the mirrors structure. First field is the host
 * name of the remote server, second is the host name of the mirror,
 * and third is the path on the mirror which corresponds to root on
 * the server.
 *
 * I try to make the client unaware that there are directories below
 * them. "cdup", and "cd .." are blocked, and the hidden part of the
 * path is stripped from the reply to a "pwd" request. I haven't
 * tested extensively though. Also I suspect bad things will happen if
 * one of the mirror servers doesn't log you in to the root directory.
 *
 * I suspect that this program would be a third the length if it were
 * written in Perl. :) */

/*
 * Basically if we read an "I" it is followed by session initialisation
 * data, a "C" and it is followed by a message from the client, and an
 * "S" and it is followed by a message from the server.
 *
 * If we write an "X" frox will forward on the message it just sent
 * us. Anything else and we are responsible for doing it ourselves. If
 * we write "S ..." we send a message to the server, and "C ....." a
 * message to the client. "L ......" sends a log message, and should be
 * followed by an action. "Q" tells frox to exit this session.
 *
 * We can't use "C" or "S" to reply to an "I", but we can reply with 
 * "R ......." where the R is followed by an IP address. Frox will redirect
 * the session to this IP.
*/

#define BLEN 1024 /*Must be greater than frox's MAX_LINE_LEN*/

struct mirror_table{
	const char *remote;
	const char *mirror;
	const char *dir;
	
	struct in_addr *r_addrs, *m_addrs;
	int n_raddrs, n_maddrs;

};
struct mirror_table mirrors[] = {
	{ "ftp.debian.org", "localhost", "/pub/mirrors/debian" },
	{ "ftp.tucows.com", "localhost", "/pub/mirrors/tucows" },
        { 0, 0, 0}
};
char *edit_cmds[] = {"SMNT", "APPE", "RNFR", "RNTO", "DELE", "RMD",
		     "MKD", "RETR", "SIZE", "STOR", "LIST", "NLST",
		     "STAT", "CWD", 0};
int edit_codes[] = {257, 150, 0};
int m_no=-1;

static int level=0, oldlevel=0, lconfirm_needed=0, inroot=1;
;

int init_mirrors(void);
int check_ip(char *buf);
void change_ip(void);
int edit_cmd(char *buf);
int edit_msg(char *buf);
int allow_path(char *arg);

int main(void)
{
	char buf[BLEN];

	if(init_mirrors()==-1) {
		printf("L CCP unable to resolve mirrors list\nQ");
		return -1;
	}

	while(fgets(buf, 1023, stdin)) {
		switch(*buf) {
		case 'C':
			if(m_no<0) printf("X\n");
			else edit_cmd(buf);
			break;
		case 'S':
			if(m_no<0) printf("X\n");
			else edit_msg(buf);
			break;
		case 'I':
			m_no = check_ip(buf);
			if(m_no<0) printf("X\n");
			else change_ip();
			break;
		}
		fflush(stdout);
	}
	return 0;
}

int init_mirrors(void)
{
	int i, j;
	char **p;
	struct hostent *hostinfo;

	for(i=0;mirrors[i].remote; i++) {
		hostinfo = gethostbyname(mirrors[i].remote);
		if (!hostinfo || hostinfo->h_addrtype != AF_INET)
			return(-1);

		for(p=hostinfo->h_addr_list, j=0; *p; p++, j++);
		mirrors[i].r_addrs=malloc(sizeof(struct in_addr) * j);
		mirrors[i].n_raddrs=j;
		for(j=0;j<mirrors[i].n_raddrs;j++) {
			mirrors[i].r_addrs[j] = *((struct in_addr *)
						hostinfo->h_addr_list[j]);
		}

		hostinfo = gethostbyname(mirrors[i].mirror);
		if (!hostinfo || hostinfo->h_addrtype != AF_INET)
			return(-1);

		for(p=hostinfo->h_addr_list, j=0; *p; p++, j++);
		mirrors[i].m_addrs=malloc(sizeof(struct in_addr) * j);
		mirrors[i].n_maddrs=j;
		for(j=0;j<mirrors[i].n_maddrs;j++) {
			mirrors[i].m_addrs[j] = *((struct in_addr *)
						hostinfo->h_addr_list[j]);
		}
	}
	return 0;
}

int check_ip(char *buf)
{
	int i, j;
	char tmp[20];
	struct in_addr match;

	sscanf(buf, "I %*s %s %*s", tmp);
	inet_aton(tmp, &match);
	for(i=0;mirrors[i].remote; i++) {
		for(j=0;j<mirrors[i].n_raddrs;j++) {
			if(match.s_addr==mirrors[i].r_addrs[j].s_addr)
				return i;
		}
	}
	return -1;
}

void change_ip(void)
{
	int j;
	j = (int) (mirrors[m_no].n_maddrs * rand()/(RAND_MAX+1.0));
	printf("L Redirecting connection for %s to local mirror\n",
	       mirrors[m_no].remote);
	printf("R %s\n", inet_ntoa(mirrors[m_no].m_addrs[j]));
}


int edit_cmd(char *buf)
{
	char *cmd, *arg, **p, tmp[BLEN];

	cmd=buf+2;
	for(arg = cmd;*arg!=' ';arg++);
	*arg++=0;
	
	arg[strlen(arg)-1]=0;

	if(!strcmp(cmd, "CDUP")) {
		if(!level) {
			printf("C 550 No such directory\n");
			return -1;
		}
		oldlevel=level--;
		lconfirm_needed=1;
		
	}

	for(p=edit_cmds;*p;p++)
		if(!strcmp(*p, cmd)) break;

	if(!allow_path(arg)) {
		printf("C 550 No such directory\n");
		return -1;
	}
	if(!strcmp(cmd, "CWD")) {
		int i = oldlevel;
		oldlevel = level;
		level = i;
		lconfirm_needed = 1;
	}
	if(!*p || (!inroot && *arg!='/') ||
	   (!strcmp(cmd, "STAT") && strlen(arg)==0) ) {
		printf("X\n");
		return 0;
	}

	strcpy(tmp, mirrors[m_no].dir);
	strcat(tmp, "/");
	strcat(tmp, arg);
	printf("S %s %s\n", cmd, tmp);
	return 1;
}

int allow_path(char *arg)
{
	int l=level;
	char *p=arg;

	if(*p=='/') {
		l=0;
		p++;
	}
	while(*p){
		if(!strncmp(p, "../", 3) || !strcmp(p, "..")) l--;
		else l++;
		if(l<0) return 0;
		for(;*p && *p!='\n' && *p!='/'; p++);
		if(!*p) break;
		while(*p++=='/');
	}
	oldlevel = l;
	return 1;
}

int edit_msg(char *buf)
{
	int *cp, i, code;
	char *cmd, *arg, *p;

	cmd=buf+2;
	for(arg = cmd;*arg!=' ';arg++);
	*arg++=0;
	code=atoi(cmd);
	
	if(lconfirm_needed) {
		if(code>299) level=oldlevel;
		else if(level) inroot=0;
		lconfirm_needed=0;
	}
	for(cp=edit_codes;*cp;cp++)
		if(*cp==code) break;
	if(!*cp) {
		printf("X\n");
		return 0;
	}

	p=strstr(arg, mirrors[m_no].dir);
	if(!p) {
		printf("X\n");
		return 0;
	}

	i=strlen(mirrors[m_no].dir);
	if(*(p+i)!='/') {
		p++;
		i--;
	}
	memmove(p, p+i, strlen(p+i)+1);

	printf("C %s %s", cmd, arg); /*Trailing \n is still in arg*/
	return 1;
}
